;; Generated package description from windower.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "windower" "0.0.1" "Helper functions for window manipulation." '((emacs "25")) :commit "520f1b59d080996a0efc2aea29a76f165d3d63f5" :authors '(("Pierre Neidhardt" . "mail@ambrevar.xyz")) :maintainer '("Pierre Neidhardt" . "mail@ambrevar.xyz") :keywords '("convenience" "tools") :url "https://gitlab.com/ambrevar/windower")
